from __future__ import annotations as __anotations
from typing import Callable as __C, Hashable as __H, Sequence as __S, Any as __A
import random
import matplotlib.pyplot as plt
from .__tdscurve import TDScurve


class TDScontainer(list[TDScurve]):

    def __or__(self, other: TDScontainer | __S[TDScurve]) -> TDScontainer:
        return TDScontainer({*self} | {*other})

    def __ror__(self, other: TDScontainer | __S[TDScurve]) -> TDScontainer:
        return TDScontainer({*other} | {*self})

    def __and__(self, other: TDScontainer | __S[TDScurve]) -> TDScontainer:
        return TDScontainer({*self} & {*other})

    def __rand__(self, other: TDScontainer | __S[TDScurve]) -> TDScontainer:
        return TDScontainer({*other} & {*self})

    def __xor__(self, other: TDScontainer | __S[TDScurve]) -> TDScontainer:
        return TDScontainer({*self} ^ {*other})

    def __rxor__(self, other: TDScontainer | __S[TDScurve]) -> TDScontainer:
        return TDScontainer({*other} ^ {*self})

    def __invert__(self) -> TDScontainer:
        return TDScontainer({*self})

    def __add__(self, other: TDScontainer | __S[TDScurve]) -> TDScontainer:
        return TDScontainer(super().__add__(self, other))

    def __radd__(self, other: TDScontainer | __S[TDScurve]) -> TDScontainer:
        return TDScontainer(super().__add__(other, self))

    def __sub__(self, other: TDScontainer | __S[TDScurve]) -> TDScontainer:
        return TDScontainer({*self} - {*other})

    def __rsub__(self, other: TDScontainer | __S[TDScurve]) -> TDScontainer:
        return TDScontainer({*other} - {*self})

    def __mul__(self, other: int) -> TDScontainer:
        return TDScontainer(super().__mul__(int(other)))

    def __mod__(self, other: TDScurve) -> TDScontainer:
        return TDScontainer({*self} - {other})

    def __pow__(self, other: __S) -> tuple[tuple[TDScurve, __A]]:
        return tuple((a, b) for a in self for b in other)

    def __rpow__(self, other: __S) -> tuple[tuple[__A, TDScurve]]:
        return tuple((a, b) for a in other for b in self)

    def __matmul__(self, func: __C[[TDScontainer], __A]) -> __A:
        return func(self)

    def __rshift__(self, func: __C[[TDScurve], __A]) -> list:
        return [func(curve) for curve in self]

    def __getitem__(self, index: int) -> TDScurve:
        return TDScontainer(super().__getitem__(index)) if isinstance(index, slice) else super().__getitem__(int(index))

    def __setitem__(self, index: int, value: TDScurve) -> None:
        super().__setitem__(index, value) if isinstance(index, slice) else super().__setitem__(int(index), value)

    def __repr__(self) -> str:
        n = self @ abs
        return F'[TDScontainer of {self @ len} TDScurve ({n} trial{"" if n <= 1 else "s"})]'

    def __abs__(self) -> int:
        return sum(x.trials_count for x in self)

    def copy(self) -> TDScontainer:
        return TDScontainer(super().copy())

    def map(self, func: __C[[TDScurve], TDScurve]) -> TDScontainer:
        return TDScontainer(curve @ func for curve in self)

    def filter(self, func: __C[[TDScurve], bool]) -> TDScontainer:
        return TDScontainer(curve for curve in self if curve @ func)

    def groupBy(self, func: str | __C[[TDScurve], __H]) -> dict[__H, TDScontainer]:
        result: dict[__H, TDScontainer] = {}

        def group_func(x): return TDScurve.getMeta(x, func) if isinstance(func, str) else func(x)
        for curve in self:
            key = curve @ group_func
            if key not in result:
                result[key] = TDScontainer()
            result[key].append(curve)

        return result

    def orderBy(self, func: str | __C[[TDScurve], __A], desc=False) -> TDScontainer:
        def order_func(x): return TDScurve.getMeta(x, func) if isinstance(func, str) else func(x)
        self.sort(key=order_func, reverse=desc)
        return self

    def getMeta(self, key: str) -> list:
        return [curve.getMeta(key) for curve in self]

    def setMeta(self, key: str, value: __S) -> TDScontainer:
        for pair in zip(self, value):
            pair[0].setMeta(key, pair[1])
        return self

    def merge(self) -> TDScurve:
        return TDScurve.sum(self)

    def mergeAs(self, name: str) -> TDScurve:
        return self.merge().setName(name)

    def bootstrap(self, size: int = None) -> TDScontainer:
        return TDScontainer(random.choices(self, k=size or self @ len))

    def distance(self, distance_func: __C[[TDScurve, TDScurve], float] = TDScurve.distance) -> list[float]:
        merged = self.merge()
        return self >> (lambda x: distance_func(x, merged))

    def boxPlot(self, size: tuple[float] = (6, 4), whis: float = 1.5, point_size: int = 10, x_position: float = 0.99,
                auto_adjust: bool = True, show: bool = True) -> TDScontainer:
        fig = plt.figure(figsize=size)
        fig.suptitle("Box Plot")
        ax: plt.Axes = fig.add_subplot(ylabel="Distance")
        ax.boxplot(self.distance(), sym="", whis=whis)
        auto_adjust or ax.set_ylim(0, 1)
        ax.scatter((x_position,) * len(self), self.distance(), s=point_size)
        show and plt.show()

        return self
